from unittest.mock import patch, call, MagicMock
from arkindex.project.tests import FixtureTestCase
from arkindex.dataimport.filetypes import \
    FileType, IIIFFileType, SurfaceFileType, ActsListFileType, MetadataFileType, ConfigFileType
from arkindex.dataimport.git import GitFlow, SimpleDiff, DiffType
from arkindex.dataimport.config import ConfigFile, ImportType, VolumesImportFormat
from arkindex.dataimport.models import DataImportMode
from arkindex.documents.models import ElementType, MetaData, MetaType, DateType, InterpretedDate


class TestFileTypes(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = cls.corpus.repos.get()
        cls.rev = cls.repo.revisions.get()
        cls.di = cls.corpus.imports.create(
            mode=DataImportMode.Repository,
            creator=cls.user,
            revision=cls.rev,
            payload={
                'repo_id': str(cls.repo.id),
                'sha': cls.rev.hash,
            },
        )
        with patch('arkindex.dataimport.git.os.path.isdir', lambda _: True):
            cls.flow = GitFlow(cls.di, 'the/')
        cls.flow.config = MagicMock(spec=ConfigFile)

    def tearDown(self):
        super().tearDown()
        self.flow.config.reset_mock()

    @patch('arkindex.dataimport.filetypes.file_types', [])
    def test_filetype_get_fail(self):
        with self.assertRaisesRegex(ValueError, 'Unknown file type'):
            FileType.get(self.flow, SimpleDiff(DiffType.Addition, 'a', 'a'))

    def test_manifest_match(self):
        self.flow.config.path_match.return_value = True
        self.flow.config.volumes_format = VolumesImportFormat.IIIF
        self.assertTrue(IIIFFileType.match('test.json', self.flow.config))

        self.assertEqual(self.flow.config.path_match.call_count, 1)
        self.assertEqual(self.flow.config.path_match.call_args, call(ImportType.Volumes, 'test.json'))

        self.assertFalse(IIIFFileType.match('test.xml', self.flow.config))

        self.flow.config.path_match.return_value = False
        self.assertFalse(IIIFFileType.match('test.json', self.flow.config))

    def test_surface_match(self):
        self.flow.config.path_match.return_value = True
        self.assertTrue(SurfaceFileType.match('path/to/some.xml', self.flow.config))
        self.assertFalse(SurfaceFileType.match('wrong/type.csv', self.flow.config))

        self.assertEqual(self.flow.config.path_match.call_count, 2)
        self.assertEqual(self.flow.config.path_match.call_args_list, [
            call(ImportType.Surfaces, 'path/to/some.xml'),
            call(ImportType.Surfaces, 'wrong/type.csv'),
        ])

        self.flow.config.path_match.return_value = False
        self.assertFalse(SurfaceFileType.match('path/to/some.xml', self.flow.config))

    @patch('arkindex.dataimport.filetypes.SurfaceImporter')
    def test_surface_handle(self, importer_mock):
        self.flow.config.path_match.return_value = True
        surface_diff = SimpleDiff(DiffType.Addition, 'a.xml', 'b.xml')
        ft = SurfaceFileType(self.flow, surface_diff)
        ft.handle()
        self.assertEqual(importer_mock.call_count, 1)
        self.assertEqual(
            importer_mock.call_args,
            call(xmlpath='the/repo/b.xml', basepath='the/repo', corpus=self.corpus),
        )
        self.assertEqual(importer_mock().run.call_count, 1)
        self.assertEqual(importer_mock().run.call_args, call())

    def test_acts_match(self):
        self.flow.config.path_match.return_value = True
        self.assertTrue(ActsListFileType.match('path/to/some.csv', self.flow.config))
        self.assertFalse(ActsListFileType.match('wrong/type.xml', self.flow.config))

        self.assertEqual(self.flow.config.path_match.call_count, 2)
        self.assertEqual(self.flow.config.path_match.call_args_list, [
            call(ImportType.Acts, 'path/to/some.csv'),
            call(ImportType.Acts, 'wrong/type.xml'),
        ])

        self.flow.config.path_match.return_value = False
        self.assertFalse(ActsListFileType.match('path/to/some.csv', self.flow.config))

    @patch('arkindex.dataimport.filetypes.CorpusSurfaceLinker')
    def test_acts_handle(self, linker_mock):
        self.flow.config.path_match.return_value = True
        acts_diff = SimpleDiff(DiffType.Modification, 'a.csv', 'b.csv')
        ft = ActsListFileType(self.flow, acts_diff)
        ft.handle()
        self.assertEqual(linker_mock.call_count, 1)
        self.assertEqual(linker_mock.call_args, call('the/repo/b.csv', self.corpus))
        self.assertEqual(linker_mock().run.call_count, 1)
        self.assertEqual(linker_mock().run.call_args, ())

    def test_config_match(self):
        self.assertTrue(ConfigFileType.match('path/to/.arkindex.yml', self.flow.config))
        self.assertFalse(ConfigFileType.match('path/to/somethingelse.yml', self.flow.config))

    def test_metadata_match(self):
        self.flow.config.path_match.return_value = True
        self.assertTrue(MetadataFileType.match('path/to/some.xml', self.flow.config))
        self.assertFalse(MetadataFileType.match('wrong/type.csv', self.flow.config))

        self.assertEqual(self.flow.config.path_match.call_count, 2)
        self.assertEqual(self.flow.config.path_match.call_args_list, [
            call(ImportType.Metadata, 'path/to/some.xml'),
            call(ImportType.Metadata, 'wrong/type.csv'),
        ])

        self.flow.config.path_match.return_value = False
        self.assertFalse(MetadataFileType.match('path/to/some.xml', self.flow.config))

    @patch('arkindex.documents.tei.Date')
    @patch('arkindex.dataimport.filetypes.TeiParser')
    def test_metadata_handle(self, tei_parser_mock, tei_elt_mock):
        db_elt = self.corpus.elements.get(type=ElementType.Act, name='Act 1')
        metadata = MetaData.objects.create(
            name='date',
            type=MetaType.Date,
            value='1333',
            element=db_elt
        )
        InterpretedDate.objects.create(
            year='1332',
            metadata=metadata,
            type=DateType.Exact
            )
        tei_parser_mock().match_database.return_value = [(db_elt, tei_elt_mock())]
        tei_elt_mock().save.return_value = [metadata]
        self.flow.config.path_match.return_value = True
        metadatas_diff = SimpleDiff(DiffType.Modification, 'a.xml', 'b.xml')
        ft = MetadataFileType(self.flow, metadatas_diff)
        ft.handle()
        self.assertEqual(tei_parser_mock.call_args, call('the/repo/b.xml'))
        self.assertEqual(tei_parser_mock().match_database.call_args, call(self.corpus))
        self.assertEqual(tei_elt_mock().save.call_count, 1)
        self.assertEqual(metadata.dates.first().year, 1333)
