from rest_framework.generics import \
    ListAPIView, ListCreateAPIView, RetrieveAPIView, RetrieveUpdateAPIView, RetrieveUpdateDestroyAPIView
from arkindex.project.pagination import PageNumberPagination
from rest_framework.exceptions import ValidationError
from django.db.models import prefetch_related_objects
from arkindex.documents.serializers.light import PageLightSerializer
from arkindex.documents.serializers.elements import (
    ElementSerializer, ElementLightSerializer, CorpusSerializer,
    PageSerializer, ActSerializer, SurfaceSerializer,
)
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.documents.models import (
    Element, ElementType, Page, Act, TranscriptionType, Transcription,
    Corpus, Right,
)
from arkindex.project.mixins import CorpusACLMixin
from arkindex.project.permissions import IsVerifiedOrReadOnly


class ElementsList(CorpusACLMixin, ListAPIView):
    """
    List documents using pagination
    """
    serializer_class = ElementLightSerializer

    def get_queryset(self):
        filters = {
            'type': self.request.query_params.get('type', ElementType.Volume),
        }
        if 'corpus' in self.request.query_params:
            filters['corpus'] = self.get_corpus(self.request.query_params['corpus'])
        else:
            filters['corpus__in'] = Corpus.objects.readable(self.request.user)
        return Element.objects.filter(**filters).prefetch_related('corpus').order_by('corpus', 'type', 'name')


class ElementRetrieve(RetrieveUpdateAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer

    def get_queryset(self):
        if self.request.method == 'GET':
            corpora = Corpus.objects.readable(self.request.user)
        else:
            corpora = Corpus.objects.writable(self.request.user)
        return Element.objects.filter(corpus__in=corpora).select_related('corpus')


class RelatedElementsList(ListAPIView):
    """
    List all documents linked to a specified document
    """
    serializer_class = ElementLightSerializer

    def get_queryset(self):
        filtering = {
            'corpus__in': Corpus.objects.readable(self.request.user),
        }
        type_param = self.request.query_params.get('type')
        if type_param is not None:
            filtering['type'] = type_param
        related = Element.objects.get_related(self.kwargs['pk'], **filtering)
        prefetch_related_objects(related, 'corpus')
        return related


class ElementPages(ListAPIView):
    """
    List pages inside an element
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.get_descending(
            self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user),
            zone__isnull=False,
        ).prefetch_related('zone__image__server')


class ElementSurfaces(ListAPIView):
    """
    List surfaces inside a volume
    """
    serializer_class = ElementLightSerializer

    def get_queryset(self):
        return Element.objects.get_descending(
            self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user),
        ).filter(type=ElementType.Surface)


class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)


class CorpusRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if request.method != 'GET' and Right.Write not in obj.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')


class CorpusPages(CorpusACLMixin, ListAPIView):
    """
    List all pages in a corpus
    """
    serializer_class = PageLightSerializer

    def get_queryset(self):
        return Page.objects.filter(corpus=self.get_corpus(self.kwargs['pk'])).select_related('zone__image__server')


class PageDetails(RetrieveAPIView):
    """
    Get details for a specific page
    """
    serializer_class = PageSerializer

    def get_queryset(self):
        return Page.objects \
                   .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
                   .select_related('zone__image__server', 'corpus') \
                   .prefetch_related('classifications__source', 'transcriptions__source')

    def get_object(self):
        page = super().get_object()
        page.parent_elements = Element.objects.get_ascending_paths(page.id)
        return page


class SurfaceDetails(RetrieveAPIView):
    """
    Get details for a specific surface
    """
    serializer_class = SurfaceSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Surface,
            corpus__in=Corpus.objects.readable(self.request.user),
        ).select_related('zone__image__server')

    def get_object(self):
        surf = super().get_object()
        surf.page = Page.objects.filter(zone__image=surf.zone.image) \
                                .select_related('zone__image__server').first()
        return surf


# TODO: Edition disabled
class ActEdit(RetrieveAPIView):
    """
    Edit acts and associated ElementPaths
    """
    serializer_class = ActSerializer

    def get_queryset(self):
        return Act.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def get_object(self):
        act = super().get_object()
        act.parent_elements = Element.objects.get_ascending_paths(act.id)
        act.child_elements = Element.objects.get_descending(act.id)
        return act


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


class ElementTranscriptions(ListAPIView):
    """
    Get transcriptions corresponding to a specific element
    Transcriptions can be filtered by type
    """
    serializer_class = TranscriptionSerializer
    pagination_class = TranscriptionsPagination

    def get_queryset(self):
        queryset = Transcription.objects.filter(
            element_id=self.kwargs['pk'],
            element__corpus__in=Corpus.objects.readable(self.request.user),
        ).prefetch_related('zone__image__server', 'source').order_by('id')
        req_type = self.request.query_params.get('type')
        if req_type:
            try:
                req_type = TranscriptionType(req_type)
                queryset = queryset.filter(type=req_type)
            except ValueError:
                raise ValidationError({'type': 'Not a valid transcription type'})
        return queryset
