from django.conf import settings
from rest_framework import status
from rest_framework.generics import CreateAPIView, UpdateAPIView
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from arkindex.documents.models import Classification, DataSource, Transcription, TranscriptionType
from arkindex.documents.serializers.ml import \
    ClassificationsSerializer, TranscriptionsSerializer, TranscriptionCreateSerializer
from arkindex.documents.indexer import Indexer
from arkindex.images.models import Zone
from arkindex.images.importer import build_transcriptions, save_transcriptions, index_transcriptions
from arkindex.project.permissions import IsVerified
from arkindex.project.polygon import Polygon


class TranscriptionCreate(CreateAPIView):
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsVerified, )

    def perform_create(self, serializer):
        transcription_type = serializer.validated_data['type']
        if transcription_type not in TranscriptionType:
            raise ValidationError("This endpoint can only import transcriptions.")
        element = serializer.validated_data['element']

        # Create a zone on the page's image
        polygon = Polygon(serializer.validated_data['polygon'])
        try:
            ts_zone, _ = Zone.objects.get_or_create(
                image=element.zone.image,
                polygon=polygon,
            )
        except Zone.MultipleObjectsReturned:
            # This should not happend, but we cannot set a unique
            # on polygon PG fields
            ts_zone = Zone.objects.filter(
                image=element.zone.image,
                polygon=polygon,
            ).first()

        ts, created = Transcription.objects.get_or_create(
            element=element,
            source=serializer.validated_data['source'],
            type=transcription_type,
            zone=ts_zone,
            text=serializer.validated_data['text'],
            defaults={
                'score': serializer.validated_data['score']
            }
        )
        if not created or ts.score != serializer.validated_data['score']:
            ts.score = serializer.validated_data['score']
            ts.save()

        # Index in ES
        Indexer().run_index(
            settings.ES_INDEX_TRANSCRIPTIONS,
            Transcription.INDEX_TYPE,
            Transcription.objects.filter(id=ts.id)
        )
        return ts

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response({'id': obj.id}, status=status.HTTP_201_CREATED, headers=headers)


class TranscriptionBulk(CreateAPIView, UpdateAPIView):
    '''
    Create transcriptions in bulk, all linked to the same image
    and parent element
    '''
    serializer_class = TranscriptionsSerializer
    permission_classes = (IsVerified, )

    def get_object(self):  # Ignore the get_object part in UpdateAPIView.update
        return

    def perform_create(self, serializer):
        self.run(serializer, delete=False)

    def perform_update(self, serializer):
        self.run(serializer, delete=True)

    def run(self, serializer, delete=False):
        parent = serializer.validated_data['parent']
        source = DataSource.from_ml_tool(serializer.validated_data['recognizer'])
        trpolygons = build_transcriptions(
            image=parent.zone.image,
            parent=parent,
            source=source,
            items=[
                {
                    'polygon': tr['polygon'],
                    'text': tr['text'],
                    'score': tr['score'],
                    'type': tr['type'],
                }
                for tr in serializer.validated_data['transcriptions']
            ],
        )
        if not trpolygons:
            return

        if delete:
            # Clear all previous transcriptions from source
            parent.transcriptions.filter(source=source).delete()

        transcriptions, _ = save_transcriptions(*trpolygons)

        # Build the Page transcription
        text = '\n'.join(
            t.text
            for t in sorted(
                parent.transcriptions.filter(
                    type=TranscriptionType.Line,
                    source=source,
                ).prefetch_related('zone'),
                key=lambda t: (t.zone.polygon.center.y, t.zone.polygon.center.x),
            )
        )

        # Add the (page transcription ID, element ID) tuple to the transcriptions tuple
        transcriptions += (
            (
                parent.transcriptions.create(
                    type=TranscriptionType.Page,
                    text=text,
                    zone=parent.zone,
                    source=source,
                ).id,
                parent.id,
            ),
        )
        index_transcriptions(transcriptions)


class ClassificationBulk(CreateAPIView):
    """
    Create multiple classifications at once for a single element from a single ML tool
    """
    serializer_class = ClassificationsSerializer
    permission_classes = (IsVerified, )

    def perform_create(self, serializer):
        parent = serializer.validated_data['parent']
        source = DataSource.from_ml_tool(serializer.validated_data['classifier'])
        parent.classifications.filter(source=source).delete()
        Classification.objects.bulk_create([
            Classification(
                page=parent,
                source=source,
                class_name=cl['class_name'],
                confidence=cl['confidence'],
            )
            for cl in serializer.validated_data['classifications']
        ])
