from django.conf import settings
from django.core.exceptions import ValidationError
from arkindex.documents.models import DateType, InterpretedDate, Act, Page
from arkindex.documents.indexer import Indexer
import unicodedata
import re
import logging

logger = logging.getLogger(__name__)

MONTHS = {
    'en': (
        'january',
        'february',
        'march',
        'april',
        'may',
        'june',
        'july',
        'august',
        'september',
        'october',
        'november',
        'december',
        ),
    'fr': (
        'janvier',
        'fevrier',
        'mars',
        'avril',
        'mai',
        'juin',
        'juillet',
        'aout',
        'septembre',
        'octobre',
        'novembre',
        'decembre',
    ),
}


def year(raw_date):
    """
    Matches 4-digit year only: '1320', '2010'
    """
    match = re.match(r'^(?P<year>\d{4})$', raw_date)
    return match.groupdict() if match else None


def year_month(raw_date):
    """
    Matches 4-digit year followed by 1 or 2-digit month: '1320-3', '2010-02'
    """
    match = re.match(r'^(?P<year>\d{4})-(?P<month>\d{1,2})$', raw_date)
    return match.groupdict() if match else None


def year_month_day(raw_date):
    """
    Matches 4-digit year followed by 1 or 2-digit month then
    1 or 2-digit day : '1320-3-12', '2010-02-31'
    """
    match = re.match(r'^(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})$', raw_date)
    return match.groupdict() if match else None


def year_month_str(raw_date):
    """
    Matches a date with alpha-written month and 4-digit year
    Both must be unique : '1320, march', 'february 2010'
    Days are not interpreted yet if present in the raw date text
    """
    years = []
    months = []
    for match in re.finditer(r'(?=(^|[^\d])(?P<year>\d{4})($|\D))', raw_date):
        years.append(match.groupdict()['year'])
    words = re.findall(r'[a-z]+', raw_date)
    for w in words:
        for month_lang in MONTHS:
            if w in MONTHS[month_lang]:
                months.append(MONTHS[month_lang].index(w) + 1)
                break
    # Exactly one year and 1 month are found
    if len(years) == len(months) == 1:
        return {'year': years[0], 'month': months[0]}
    return None


DATE_FUNCTIONS_TABLE = (
    year,
    year_month,
    year_month_day,
    year_month_str,
)


def _normalize(raw_date):
    """
    Return an unaccented lowercase string from input
    """
    return unicodedata.normalize('NFKD', raw_date) \
        .encode('ASCII', 'ignore') \
        .lower() \
        .decode("utf-8")


def parse_date(raw_date, functions_table=DATE_FUNCTIONS_TABLE):
    """
    Return a list of interpreted dates from a raw date string
    """
    assert isinstance(raw_date, str)
    # Try to match regex one by one
    for f in functions_table:
        try:
            date_elts = f(_normalize(raw_date))
            if date_elts:
                assert isinstance(date_elts, dict)
                date = InterpretedDate(type=DateType.Exact)
                for k in date_elts:
                    setattr(date, k, int(date_elts[k]))
                try:
                    date.full_clean(exclude=('metadata', ))
                except ValidationError as e:
                    logger.warning('Date fields are incorrect : {}'.format(e))
                    continue
                return (date, )
        except Exception:
            logger.warning('Failed parsing {} with function {}'.format(raw_date, f.__name__))
    logger.warning('Date not supported : {}'.format(raw_date))
    return ()


class DateParser(object):
    """
    Interprete date elements from non-standard TEI input
    Input examples :  "1325-11-04"
                      "s.d. [1323-1325]"
    """

    def __init__(self, metadatas):
        self.metadatas = metadatas.prefetch_related('dates')

    def index(self):
        """
        Create or Update ElasticSearch index for indexed elements
        """
        acts = Act.objects.filter(metadatas__in=self.metadatas)
        if acts.exists():
            Indexer().run_index(settings.ES_INDEX_ACTS, Act.INDEX_TYPE, acts)
        pages = Page.objects.filter(metadatas__in=self.metadatas)
        if pages.exists():
            Indexer().run_index(settings.ES_INDEX_PAGES, Page.INDEX_TYPE, pages)

    def delete(self):
        """
        Delete a queryset of interpreted dates corresponding to a metadatas list
        """
        dates = InterpretedDate.objects.filter(metadata__in=self.metadatas)
        nb_deleted, _ = dates.delete()
        logger.info('Deleted {} old interpreted dates in database'.format(nb_deleted))

    def run(self, dry_run=False):
        """
        Create and save new interpreted dates for a list of metadatas
        Optionnaly Delete old interpreted dates
        """
        old_coverage = self.metadatas.filter(dates__isnull=False).count()
        new_coverage = 0
        if not dry_run:
            self.delete()
        for md in self.metadatas:
            dates = parse_date(md.value)
            if dates:
                new_coverage += 1
            if not dry_run:
                for date in dates:
                    date.metadata = md
                    date.save()
        logger.info('{} interpreted dates have been interpreted'.format(new_coverage))
        total_metadatas = self.metadatas.count()
        if total_metadatas:
            logger.info('Coverage of Date Metadatas is now {:.2%} ({:+.2%})'.format(
                new_coverage / total_metadatas,
                (new_coverage - old_coverage) / total_metadatas
            ))
        if not dry_run:
            self.index()
