from django.db import models, transaction
from django.contrib.postgres.indexes import GinIndex
from django.utils.functional import cached_property
from django.core.validators import MinValueValidator, MaxValueValidator
from django.core.exceptions import ValidationError
from enumfields import EnumField, Enum
from arkindex_common.ml_tool import MLToolType
from arkindex.project.models import IndexableModel
from arkindex.project.fields import ArrayField
from arkindex.documents.managers import ElementManager, CorpusManager
from datetime import datetime
import uuid
import logging
import itertools

logger = logging.getLogger(__name__)

# Language used for display
DEFAULT_LANGUAGE = 'en'


class Right(Enum):
    Read = 'read'
    Write = 'write'
    Admin = 'admin'


class Corpus(models.Model):
    '''
    A group of elements, usually linked to users or projects
    '''
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250)
    description = models.TextField(default="")

    # Is this corpus publicly readable ?
    public = models.BooleanField(default=False)

    # Specific manager for ACL
    objects = CorpusManager()

    class Meta:
        verbose_name_plural = 'corpora'

    def __str__(self):
        return self.name

    def get_acl_rights(self, user):
        '''
        List Access rights on this corpus for this user
        '''
        from arkindex.users.models import CorpusRight

        # Read only for public + anon
        if user.is_anonymous:
            return self.public and [Right.Read] or []

        # Full rights for admins and internal users
        if user.is_admin or user.is_internal:
            return [Right.Read, Right.Write, Right.Admin]

        # Check acl
        try:
            access = self.corpus_right.get(user=user)
            rights = [Right.Read, ]
            if access.can_write or access.can_admin:
                rights.append(Right.Write)
            if access.can_admin:
                rights.append(Right.Admin)
        except CorpusRight.DoesNotExist:
            # Default to RO for public
            rights = self.public and [Right.Read] or []

        return rights


class ElementType(Enum):
    Volume = 'volume'
    Register = 'register'
    Act = 'act'
    Page = 'page'
    Surface = 'surface'


class ElementPath(models.Model):
    """
    Rolling cache table, with full paths exposed for each element
    An element can have multiple paths up to the top
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    element = models.ForeignKey('documents.Element', related_name='paths', on_delete=models.CASCADE)
    path = ArrayField(
        models.UUIDField(),
    )
    ordering = models.PositiveIntegerField(default=0)

    class Meta:
        indexes = [
            GinIndex(fields=['path']),
        ]


class Element(IndexableModel):
    """
    aka Unité Codicologique
    """
    corpus = models.ForeignKey('documents.Corpus', related_name='elements', on_delete=models.CASCADE)
    type = EnumField(ElementType, max_length=50, db_index=True)
    name = models.CharField(max_length=250)
    zone = models.ForeignKey(
        'images.Zone',
        on_delete=models.SET_NULL,
        related_name='elements',
        null=True,
        blank=True,
    )

    objects = ElementManager()

    def get_thumbnail(self):
        """Get an Image instance corresponding to a thumbnail for this element."""
        if self.type == ElementType.Volume:
            first_page = Page.objects.get_descending(self.id).first()
            return first_page and first_page.zone.image
        elif self.type == ElementType.Page:
            return self.page.images.first()
        return None

    @transaction.atomic
    def add_parent(self, parent, order=None):
        '''
        Add an element as ancestor
        '''
        assert isinstance(parent, Element)

        # Check link does not already exist
        current_links = ElementPath.objects.filter(element=self, path__last=parent.id)
        if current_links.exists():
            if order is not None:
                # Update link ordering
                current_links.update(ordering=order)
            return

        # Check for cycles
        parent_asc_paths = list(ElementPath.objects.filter(element_id=parent.id).values_list('path', flat=True))
        if self.id in itertools.chain(*parent_asc_paths):
            raise ValueError('Cannot add a descendant as a parent')

        if order is None:
            order = parent.get_next_order(self.type)

        if parent_asc_paths:
            # Add parent to all paths
            parent_paths = [
                path + [parent.id]
                for path in parent_asc_paths
            ]
        else:
            # Handle element having no parents
            parent_paths = [
                [parent.id],
            ]

        # Get ascending paths of all children
        children_ids = list(Element.objects.get_descending(self.id).values_list('id', flat=True))
        children_epaths_qs = ElementPath.objects.filter(element_id__in=children_ids, path__contains=[self.id])
        children_epaths = list(children_epaths_qs)

        # Delete the old paths
        children_epaths_qs.delete()

        # Use (element, path) tuples to ensure paths are unique
        new_epaths = {
            (self, tuple(path)): ElementPath(element=self, path=path, ordering=order or 0)
            for path in parent_paths
        }

        for epath in children_epaths:
            base_path = epath.path

            # This is a path with another parent, keep it
            if epath.path[0] != self.id:
                new_epaths[epath.element, tuple(epath.path)] = epath
                base_path = epath.path[epath.path.index(self.id):]  # Strip the other parent

            # For each of my parents
            for parent_asc_path in parent_paths:
                # Build new path with my parents in it
                new_path = parent_asc_path + base_path

                new_epaths[epath.element, tuple(new_path)] = ElementPath(
                    element=epath.element,
                    path=new_path,
                    ordering=epath.ordering,
                )

        # Finally create all new paths
        ElementPath.objects.bulk_create(new_epaths.values())

    def get_next_order(self, type):
        """
        Find the next ordering for a new child in ElementPath
        """
        assert isinstance(type, ElementType)
        return ElementPath.objects \
                          .filter(path__last=self.id, element__type=type) \
                          .aggregate(max=models.Max('ordering') + 1)['max'] or 0

    @transaction.atomic
    def remove_child(self, child):
        '''
        Dissociate a child element
        '''
        assert isinstance(child, Element)

        # Check child is a direct child of this element
        if not ElementPath.objects.filter(element=child, path__last=self.id).exists():
            return

        new_epaths = {}

        # If child doesn't have any other parent, updates are needed
        if not ElementPath.objects.filter(element=child).exclude(path__last=self.id).exists():

            # Get all the child's children's paths
            children_epaths = list(ElementPath.objects.filter(
                element__in=list(Element.objects.get_descending(child.id)),
                path__contains=[self.id]
            ))

            for epath in children_epaths:
                new_path = epath.path[epath.path.index(self.id) + 1:]  # Strip this parent's path

                new_epaths[epath.element, tuple(new_path)] = ElementPath(
                    element=epath.element,
                    path=new_path,
                    ordering=epath.ordering,
                )

        # Delete everything
        ElementPath.objects.filter(
            element__in=list(Element.objects.get_descending(child.id)) + [child],
            path__contains=[self.id]
        ).delete()

        # Recreate paths as needed
        ElementPath.objects.bulk_create(new_epaths.values())

    @cached_property
    def thumbnail(self):
        from arkindex.images.models import Thumbnail  # Prevent circular imports
        return Thumbnail(self)

    def generate_thumbnail(self, **kwargs):
        '''
        Build a thumbnail for this element
        '''
        logger.info("Creating thumbnail for element {}".format(self))
        self.thumbnail.create(**kwargs)

    def __str__(self):
        return '{} : {}'.format(self.type, self.name)


class PageType(Enum):
    Page = 'page'
    Summary = 'summary'
    Garde = 'garde'
    ContreGarde = 'contregarde'
    ContrePlat = 'contreplat'
    Plat = 'plat'
    Copy = 'copy'


class PageDirection(Enum):
    Recto = 'recto'
    Verso = 'verso'


class PageComplement(Enum):
    Bis = 'bis'
    Ter = 'ter'
    Quat = 'quat'


class Page(Element):
    """
    with folio numbering
    """
    INDEX_TYPE = 'page'

    folio = models.CharField(max_length=250)

    # Parsed folio
    page_type = EnumField(PageType, max_length=50, null=True, blank=True)
    nb = models.PositiveIntegerField(null=True, blank=True)
    direction = EnumField(PageDirection, max_length=10, null=True, blank=True)
    complement = EnumField(PageComplement, max_length=10, null=True, blank=True)
    objects = ElementManager()

    @property
    def display_name(self):
        """
        Generate a pretty display name for a page.
        If the folio parsing had failed, will return the raw folio.
        """
        if self.page_type is None and self.nb is None and self.direction is None:
            if self.folio.strip() == "":
                return "Unknown folio"
            return self.folio

        text = self.page_type.name if self.page_type is not None else 'Page'

        if self.nb is not None:
            text += ' no. {0:d}'.format(self.nb)
        else:
            text += ', unnumbered'

        if self.direction is not None:
            text += ', {}'.format(self.direction.value)

        return text.capitalize()

    def save(self, *args, **kwargs):
        # TODO: move this in Element through introspection
        self.type = ElementType.Page
        super().save(*args, **kwargs)

    def same_as(self, other):
        """
        Python's comparison methods are used on this class to perform sorting.
        This method checks for equality as in there are no differences between the two pages.
        Used by Git imports to tell if a page has been modified in a newer revision.
        """
        # Prevent AttributeError when a Page is an Element and its subclass has not been used
        if not isinstance(other, Page):
            if not isinstance(other, Element) and hasattr(other, 'page'):
                return False
            other = other.page
        return self.zone_id == other.zone_id and self.folio == other.folio

    def __lt__(self, other):
        if self.page_type == other.page_type:
            if self.nb == other.nb:
                if self.direction == other.direction:
                    # cmp complement
                    if self.complement is None and other.complement is not None:
                        return True
                    if self.complement == PageComplement.Bis and other.complement not in (None, PageComplement.Bis):
                        return True
                    if self.complement == PageComplement.Ter and other.complement == PageComplement.Quat:
                        return True
                    return False

                # cmp direction
                return self.direction == PageDirection.Recto and other.direction == PageDirection.Verso

            # cmp page number
            return self.nb < other.nb

        # cmp page type
        if self.page_type == PageType.Summary:
            return True
        if self.page_type == PageType.Page and other.page_type != PageType.Summary:
            return True
        return False

    def __eq__(self, other):
        # Prevent AttributeError when a Page is an Element and its subclass has not been used
        if not isinstance(other, Page):
            if not isinstance(other, Element) or not hasattr(other, 'page'):
                return False
            other = other.page
        return self.page_type == other.page_type \
            and self.nb == other.nb \
            and self.direction == other.direction \
            and self.complement == other.complement

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __str__(self):
        out = [self.type.value, ]
        if self.nb:
            out.append('n°{}'.format(self.nb))
        if self.direction:
            out.append(self.direction.value)
        # Default to raw folio
        if not self.nb and not self.direction:
            out.append(self.folio)
        return ' '.join(out)

    def build_search_index(self):
        """
        Structure indexed into ElasticSearch
        """
        return {
            'corpus': self.corpus_id,
            'volumes': [v.name for v in Element.objects.get_ascending(self.id, type=ElementType.Volume)],
            'transcriptions': [
                {
                    'id': t.id,
                    'type': t.type.value,
                    'score': t.score,
                    'text': t.text,
                }
                for t in self.transcriptions.all()
            ]
        }

    @property
    def page_transcriptions(self):
        """
        A helper property to write a serializer than only uses Page transcriptions
        """
        return self.transcriptions.filter(type=TranscriptionType.Page)

    @property
    def best_classes(self):
        """
        The three most probable classifications
        """
        return self.classifications.order_by('-confidence')[:3]


class Act(Element):
    """
    A logical subdivision that can span on multiple images via zones
    """
    INDEX_TYPE = 'act'

    number = models.CharField(max_length=30)
    folio = models.CharField(max_length=250)

    def save(self, *args, **kwargs):
        # TODO: move this in Element through introspection
        self.type = ElementType.Act
        super().save(*args, **kwargs)

    def build_search_index(self):
        """
        Structure indexed into ElasticSearch
        """
        surfaces = Element.objects.get_descending(self.id).filter(type=ElementType.Surface)
        transcriptions = [
            Transcription.objects.filter(
                zone__image=s.zone.image,
                zone__polygon__in=s.zone.polygon
            )
            for s in surfaces
        ]
        interpreted_dates = itertools.chain(*[md.dates.all() for md in self.metadatas.all()])
        # Concatenate values obtained creating ranges from interpreted dates
        date_range = {k: val for date in interpreted_dates for k, val in date.to_es_range().items()}
        return {
            'corpus': self.corpus_id,
            'volumes': [v.name for v in Element.objects.get_ascending(self.id, type=ElementType.Volume)],
            'transcriptions': [
                {
                    'id': t.id,
                    'type': t.type.value,
                    'score': t.score,
                    'text': t.text,
                }
                for sublist in transcriptions for t in sublist
            ],
            'date_range': [date_range] if date_range else []
        }


class DataSource(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    type = EnumField(MLToolType)
    slug = models.CharField(max_length=100)
    revision = models.CharField(max_length=100)
    internal = models.BooleanField()

    class Meta:
        unique_together = (
            ('type', 'slug', 'revision'),
        )

    @staticmethod
    def from_ml_tool(ml_tool):
        return DataSource.objects.get_or_create(
            type=ml_tool.type,
            slug=ml_tool.slug,
            revision=str(ml_tool.version),
            internal=True,
        )[0]


class TranscriptionType(Enum):
    Page = 'page'
    Paragraph = 'paragraph'
    Line = 'line'
    Word = 'word'
    Character = 'character'


class Transcription(models.Model):
    """
    A transcription on:
     * a zone on an image
    """
    INDEX_TYPE = 'transcription'

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    element = models.ForeignKey(
        Element,
        on_delete=models.CASCADE,
        related_name='transcriptions',
    )
    type = EnumField(
        TranscriptionType,
        default=TranscriptionType.Word,
        max_length=50,
        db_index=True,
    )
    zone = models.ForeignKey(
        'images.Zone',
        on_delete=models.SET_NULL,
        related_name='transcriptions',
        null=True,
        blank=True,
    )
    source = models.ForeignKey(
        DataSource,
        on_delete=models.CASCADE,
        related_name='transcriptions',
    )
    text = models.TextField(null=True, blank=True)
    score = models.FloatField(null=True, blank=True)

    class Meta:
        # The following index was attempted with md5(text) in a manual migration
        # but it causes too many performance issues.
        # unique_together = (
        #     ('element', 'zone', 'text')
        # )
        pass

    def __str__(self):
        return 'Transcription: {}'.format(self.text[:20])

    def build_search_index(self):
        """
        Structure indexed into Elastic Search
        """
        return {
            'element': self.element_id,
            'type': self.type.value,
            'score': self.score,
            'text': self.text,
            'corpus': self.element.corpus_id,
        }


class Classification(models.Model):
    """
    A result of a classifier on a page
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name='classifications',
    )
    source = models.ForeignKey(
        DataSource,
        on_delete=models.CASCADE,
        related_name='classifications',
    )
    class_name = models.CharField(max_length=100)
    confidence = models.FloatField(null=True, blank=True)

    class Meta:
        unique_together = (
            ('page', 'source', 'class_name'),
        )


class MetaType(Enum):
    Text = 'text'
    HTML = 'html'
    Date = 'date'
    Location = 'location'


class MetaData(models.Model):
    '''
    Metadatas for elements
    '''
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    element = models.ForeignKey(Element, related_name='metadatas', on_delete=models.CASCADE)
    name = models.CharField(max_length=250)
    type = EnumField(MetaType, max_length=50, db_index=True)
    value = models.TextField()
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.SET_NULL, blank=True, null=True)
    index = models.PositiveIntegerField(default=0)

    class Meta:
        ordering = ('element', 'name', 'index')
        unique_together = (
            ('element', 'name', 'index'),
        )

    def is_single_line(self):
        return self.type != MetaType.HTML and not self.is_list() and len(self.value.splitlines()) <= 1

    def is_list(self):
        # TODO: Remove is_list when we move to Entities
        return self.name.lower() in ('persons', 'places', 'subjects')

    def __eq__(self, other):
        if not isinstance(other, MetaData):
            return False
        return (self.is_single_line(), self.type == MetaType.HTML, self.is_list(), self.name) == (
            other.is_single_line(), other.type == MetaType.HTML, other.is_list(), self.name)

    def __gt__(self, other):
        if self.is_single_line():
            return self.name.__gt__(other.name)
        if self.type == MetaType.HTML:
            return other.is_single_line()
        if self.is_list():
            return other.is_single_line() or other.type == MetaType.HTML
        return True

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)


class DateType(Enum):
    Exact = 'exact'
    Lower = 'lower'
    Upper = 'upper'
    Unknown = 'unknown'


class DatePrecision(Enum):
    Year = 'y'
    Month = 'M'
    Day = 'd'


class InterpretedDate(models.Model):
    year = models.PositiveSmallIntegerField()
    month = models.PositiveSmallIntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(12)],
        null=True,
        blank=True,
    )
    day = models.PositiveSmallIntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(31)],
        null=True,
        blank=True,
    )
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    metadata = models.ForeignKey(MetaData, related_name='dates', on_delete=models.CASCADE)
    type = EnumField(DateType, max_length=16)

    def clean(self):
        # Check if day is correct depending on year and month
        if self.precision == DatePrecision.Day:
            try:
                datetime(*tuple(self))
            except ValueError:
                raise ValidationError('Date format is incorrect {}'.format(self))
        super().clean()

    @property
    def precision(self):
        """
        Return precision of the date depending on present elements
        """
        if self.month and self.day:
            return DatePrecision.Day
        elif self.month:
            return DatePrecision.Month
        return DatePrecision.Year

    def __iter__(self):
        return iter((self.year, self.month, self.day))

    def __eq__(self, other):
        return tuple(self) == tuple(other) and self.type == other.type

    def __gt__(self, other):
        """
        Compare dates by year, month then day if comparison is possible
        """
        for s, o in zip(tuple(self), tuple(other)):
            if s is None or o is None:
                return
            if s == o:
                continue
            return s > o

    def es_round_up(self):
        """
        Return a rounded up date using ElasticSearch date math syntax depending if month/day are present
        (https://www.elastic.co/guide/en/elasticsearch/reference/6.2/common-options.html#date-math)
        Ex: '1990' date will be represented as '1999||+1y' which means '2000' for ElasticSearch
            '2000-01' will return '2000-01||+1M' ('2000-02')
        """
        return '{}||+1{}'.format(self.es_str(), self.precision.value)

    def to_es_range(self):
        """
        Return a dict containing optional ElasticSearch indexing date-range fields
            gte: greater than or equal
            lt: lower than
        Ex: Exact date 1920-01 will return {'gte': '1920-01', 'lt':'1920-01||+1M'}
            Upper bound date 1860 will return {'lt': '1860||+1y'} only
        """
        # Round date if needed
        es_range = {}
        if self.type in (DateType.Lower, DateType.Exact):
            es_range['gte'] = self.es_str()
        if self.type in (DateType.Upper, DateType.Exact):
            es_range['lt'] = self.es_round_up()
        return es_range

    def es_str(self):
        return '-'.join('{:02d}'.format(e) for e in tuple(self) if e)

    def __str__(self):
        return self.es_str()
