from arkindex.documents.models import Transcription, Act, Page, Element
from itertools import chain
import uuid


def search_transcriptions_post(data):
    '''
    Search pages with a transcription
    List pages ordered by document
    '''
    results = data['hits']['hits']
    transcription_ids = [r['_id'] for r in results]
    ts = Transcription.objects \
                      .filter(id__in=transcription_ids) \
                      .order_by('-score') \
                      .prefetch_related('zone__image__server', 'element', 'source')
    # Recompute in case some transcriptions are not found in Postgres
    transcription_ids = [t.element_id for t in ts]
    all_parent_paths = Element.objects.get_ascendings_paths(*transcription_ids)
    for trans in ts:
        trans.parent_paths = all_parent_paths.get(trans.element_id, [])
    return ts


def search_transcriptions_filter_post(data, element_id):
    if not isinstance(element_id, uuid.UUID):
        element_id = uuid.UUID(element_id)
    return filter(
        lambda t: element_id in [p.id for p in chain(*t.parent_paths)],
        search_transcriptions_post(data)
    )


def search_pages_post(data):
    """
    Search pages containing query
    """
    return search_nested_post(Page, data)


def search_acts_post(data):
    """
    Search acts containing query
    """
    return search_nested_post(Act, data)


def search_nested_post(model, data):
    """
    Search a specific element type for a query
    Returns a list of `model` instances
    """
    assert isinstance(model, type), "Model argument must be a class"
    assert issubclass(model, Element), "Model must be an Element"

    results = data['hits']['hits']
    elt_ids = [uuid.UUID(r['_id']) for r in results]
    if not elt_ids:
        return
    tr_ids = [
        hit['_source']['id']
        for result in results
        for hit in result['inner_hits']['transcriptions']['hits']['hits']
    ]
    tr_totals = {
        uuid.UUID(result['_id']): result['inner_hits']['transcriptions']['hits']['total']
        for result in results
    }

    transcriptions = {
        t.id: t
        for t in Transcription.objects.filter(id__in=tr_ids).prefetch_related('zone__image__server')
    }

    elts_tr_ids = {
        uuid.UUID(result['_id']): [
            uuid.UUID(hit['_source']['id'])
            for hit in result['inner_hits']['transcriptions']['hits']['hits']
        ] for result in results
    }

    elts_date_range = {
        uuid.UUID(result['_id']): result['_source']['date_range'][0]
        if result['_source'].get('date_range') else {}
        for result in results
    }

    elts = list(model.objects.filter(id__in=elt_ids).prefetch_related('corpus'))
    # Preserve the ordering given by ElasticSearch
    ordered_elts = list(filter(None, map(lambda eid: next((e for e in elts if e.id == eid), None), elt_ids)))

    all_paths = Element.objects.get_ascendings_paths(*(e.id for e in ordered_elts))

    for elt in ordered_elts:
        elt.transcriptions_results = list(filter(None, [transcriptions.get(tid) for tid in elts_tr_ids[elt.id]]))
        elt.total_transcriptions = tr_totals[elt.id]
        elt.parent_paths = all_paths.get(elt.id, [])
        date_range = elts_date_range.get(elt.id, {})
        elt.date_from = date_range.get('gte')
        elt.date_to = date_range['lt'].partition('||')[0] if 'lt' in date_range else None

    return ordered_elts
