from django.urls import reverse
from rest_framework import status
from arkindex.documents.models import Element, ElementType, DataSource, TranscriptionType
from arkindex.project.tests import FixtureAPITestCase


class TestElementsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = Element.objects.get(name='Volume 1', type=ElementType.Volume, corpus=cls.corpus)
        cls.page = Element.objects.filter(type=ElementType.Page, corpus=cls.corpus).first().page
        cls.src = DataSource.objects.get(slug='test')

    def test_get_element(self):
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.vol.id))
        self.assertEqual(data['name'], 'Volume 1')
        self.assertEqual(data['corpus']['id'], str(self.corpus.id))
        self.assertEqual(data['type'], 'volume')

    def test_patch_element_acl(self):
        """
        Check patching an element as anonymous user is not possible
        """
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_element(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['name'], 'Untitled (2)')
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.name, 'Untitled (2)')

    def test_get_transcriptions(self):
        self.page.transcriptions.all().delete()
        transcriptions = []
        for i in range(10):
            transcriptions.append(self.page.transcriptions.create(source_id=self.src.id, type=TranscriptionType.Word))
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 10)
        self.assertEqual(len(transcriptions), 10)
        # Wrong transcription type
        url = reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)})
        url += '?type=potato'
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
