from django.urls import path
from django.views.generic.base import RedirectView

from arkindex.documents.api.elements import (
    ElementsList, RelatedElementsList, ElementRetrieve, ElementPages, ElementSurfaces,
    CorpusList, CorpusRetrieve, CorpusPages, ActEdit, PageDetails, SurfaceDetails,
    ElementTranscriptions,
)
from arkindex.documents.api.search import PageSearch, ActSearch
from arkindex.documents.api.ml import ClassificationBulk, TranscriptionCreate, TranscriptionBulk
from arkindex.documents.api.iiif import (
    VolumeManifest, ActManifest, PageAnnotationList, PageActAnnotationList, SurfaceAnnotationList,
    TranscriptionSearchAnnotationList,
)
from arkindex.dataimport.api import (
    DataImportsList, DataImportDetails, DataImportRetry, DataImportFailures, DataImportDemo,
    DataFileList, DataFileRetrieve, DataFileUpload, DataImportFromFiles,
    RepositoryList, RepositoryRetrieve, RepositoryStartImport,
    GitRepositoryImportHook, AvailableRepositoriesList, ElementHistory, MLToolList,
)
from arkindex.users.api import (
    ProvidersList, CredentialsList, CredentialsRetrieve, OAuthSignIn, OAuthRetry, OAuthCallback,
    UserRetrieve, UserCreate, UserEmailLogin, UserEmailVerification, PasswordReset, PasswordResetConfirm,
)

api = [

    # Elements
    path('elements/<uuid:pk>/pages/', ElementPages.as_view(), name='volume-pages'),
    path('elements/<uuid:pk>/surfaces/', ElementSurfaces.as_view(), name='volume-surfaces'),
    path('elements/<uuid:pk>/', RelatedElementsList.as_view(), name='related-elements'),
    path('elements/', ElementsList.as_view(), name='elements'),
    path('element/<uuid:pk>/', ElementRetrieve.as_view(), name='element-retrieve'),
    path('element/<uuid:pk>/history/', ElementHistory.as_view(), name='element-history'),
    path('element/<uuid:pk>/transcriptions/', ElementTranscriptions.as_view(), name='element-transcriptions'),
    path('page/<uuid:pk>/', PageDetails.as_view(), name='page-details'),
    path('surface/<uuid:pk>/', SurfaceDetails.as_view(), name='surface-details'),
    path('corpus/', CorpusList.as_view(), name='corpus'),
    path('corpus/<uuid:pk>/', CorpusRetrieve.as_view(), name='corpus-retrieve'),
    path('corpus/<uuid:pk>/pages/', CorpusPages.as_view(), name='corpus-pages'),

    # Manifests
    path('manifest/<uuid:pk>/pages/', VolumeManifest.as_view(), name='volume-manifest'),
    path('manifest/<uuid:pk>/act/', ActManifest.as_view(), name='act-manifest'),
    path('manifest/<uuid:pk>/transcriptions/', PageAnnotationList.as_view(), name='page-transcription-manifest'),
    path('manifest/<uuid:pk>/acts/', PageActAnnotationList.as_view(), name='page-act-manifest'),
    path('manifest/<uuid:pk>/surface/', SurfaceAnnotationList.as_view(), name='surface-manifest'),
    path('manifest/<uuid:pk>/search/', TranscriptionSearchAnnotationList.as_view(), name='ts-search-manifest'),

    # Placeholder URLs for IIIF IDs
    path(
        'manifest/<uuid:pk>/sequence/',
        RedirectView.as_view(pattern_name='api:volume-manifest', permanent=False),
        name='sequence-manifest',
    ),
    path(
        'manifest/<uuid:pk>/canvas/',
        RedirectView.as_view(pattern_name='api:page-act-manifest', permanent=False),
        name='canvas-manifest',
    ),
    path(
        'manifest/<uuid:page_pk>/transcriptions/<uuid:transcription_pk>/',
        RedirectView.as_view(pattern_name='api:page-transcription-manifest', permanent=False),
        name='transcription-manifest',
    ),
    path(
        'manifest/<uuid:page_pk>/surfaces/<uuid:surface_pk>/',
        RedirectView.as_view(pattern_name='api:page-act-manifest', permanent=False),
        name='surface-manifest',
    ),

    # Search transcriptions
    path('pages/', PageSearch.as_view(), name='page-search'),
    path('acts/', ActSearch.as_view(), name='act-search'),

    # Edit acts
    path('act/<uuid:pk>/', ActEdit.as_view(), name='act-edit'),

    # Ingest transcriptions
    path('transcription/', TranscriptionCreate.as_view(), name='transcription-create'),
    path('transcription/bulk/', TranscriptionBulk.as_view(), name='transcription-bulk'),
    path('classification/bulk/', ClassificationBulk.as_view(), name='classification-bulk'),

    # Git import workflows
    path('imports/repos/', RepositoryList.as_view(), name='repository-list'),
    path('imports/repos/<uuid:pk>/', RepositoryRetrieve.as_view(), name='repository-retrieve'),
    path('imports/repos/<uuid:pk>/start/', RepositoryStartImport.as_view(), name='repository-import'),
    path('imports/repos/search/<uuid:pk>/', AvailableRepositoriesList.as_view(), name='available-repositories'),
    path('imports/hook/<uuid:pk>/', GitRepositoryImportHook.as_view(), name='import-hook'),

    # Import workflows
    path('imports/', DataImportsList.as_view(), name='import-list'),
    path('imports/fromfiles/', DataImportFromFiles.as_view(), name='import-from-files'),
    path('imports/mltools/', MLToolList.as_view(), name='ml-tool-list'),
    path('imports/<uuid:pk>/', DataImportDetails.as_view(), name='import-details'),
    path('imports/<uuid:pk>/retry/', DataImportRetry.as_view(), name='import-retry'),
    path('imports/<uuid:pk>/failures/', DataImportFailures.as_view(), name='import-failures'),
    path('imports/demo/<uuid:pk>/', DataImportDemo.as_view(), name='import-demo'),
    path('imports/files/<uuid:pk>/', DataFileList.as_view(), name='file-list'),
    path('imports/file/<uuid:pk>/', DataFileRetrieve.as_view(), name='file-retrieve'),
    path('imports/upload/<uuid:pk>/', DataFileUpload.as_view(), name='file-upload'),

    # Manage OAuth integrations
    path('oauth/providers/', ProvidersList.as_view(), name='providers-list'),
    path('oauth/providers/<provider>/signin/', OAuthSignIn.as_view(), name='oauth-signin'),
    path('oauth/providers/<provider>/callback/', OAuthCallback.as_view(), name='oauth-callback'),
    path('oauth/credentials/', CredentialsList.as_view(), name='credentials-list'),
    path('oauth/credentials/<uuid:pk>/', CredentialsRetrieve.as_view(), name='credentials-retrieve'),
    path('oauth/credentials/<uuid:pk>/retry/', OAuthRetry.as_view(), name='oauth-retry'),

    # Authentication
    path('user/', UserRetrieve.as_view(), name='user-retrieve'),
    path('user/new/', UserCreate.as_view(), name='user-new'),
    path('user/login/', UserEmailLogin.as_view(), name='user-login'),
    path('user/token/', UserEmailVerification.as_view(), name='user-token'),
    path('user/password-reset/', PasswordReset.as_view(), name='password-reset'),
    path('user/password-reset/confirm/', PasswordResetConfirm.as_view(), name='password-reset-confirm'),
]
