from django.conf import settings
from rest_framework import serializers
from enum import Enum
from arkindex_common.ml_tool import MLTool, MLToolType
from arkindex.project.tools import elasticsearch_escape


class EnumField(serializers.ChoiceField):
    """
    Serializes an enum field into a JSON value
    """

    def __init__(self, enum, *args, **kwargs):
        assert issubclass(enum, Enum) or not enum
        self.enum = enum
        choices = [(item.value, item.name) for item in self.enum]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        return obj.value

    def to_internal_value(self, data):
        assert self.enum is not None, "No enum set on EnumField"
        try:
            return self.enum(data)
        except ValueError:
            raise serializers.ValidationError('Value is not of type {}'.format(self.enum.__name__))


class MLToolField(serializers.CharField):

    def __init__(self, tool_type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(tool_type, MLToolType)
        self.tool_type = tool_type

    def to_representation(self, obj):
        return obj.slug

    def to_internal_value(self, data):
        try:
            return MLTool.get(settings.ML_CLASSIFIERS_DIR, self.tool_type, data)
        except ValueError as e:
            raise serializers.ValidationError(str(e))


class SearchTermsField(serializers.CharField):
    """
    Serialize ElastiSearch query terms
    """
    def to_internal_value(self, query_terms):
        return elasticsearch_escape(query_terms)
